@@ -91,7 +91,7 @@
   :type 'list)
 
 (defcustom lsdb-interesting-header-alist
-  '(("Organization" nil organization)
+  `(("Organization" nil organization)
     ("\\(X-\\)?User-Agent\\|X-Mailer\\|X-Newsreader" nil user-agent)
     ("\\(X-\\)?ML-Name" nil mailing-list)
     ("List-Id" "\\(.*\\)[ \t]+<[^>]+>\\'" mailing-list "\\1")
@@ -99,7 +99,9 @@
     ("Delivered-To" "mailing list[ \t]+\\([^@]+\\)@.*" mailing-list "\\1")
     ("\\(X-URL\\|X-URI\\)" nil www)
     ("X-Attribution\\|X-cite-me" nil attribution)
-    ("X-Face" nil x-face))
+    ("X-Face" nil x-face)
+    ("Face" nil face)
+    (,lsdb-sender-headers nil sender))
   "Alist of headers we are interested in.
 The format of elements of this list should be
      (FIELD-NAME REGEXP ENTRY STRING)
@@ -117,7 +119,9 @@ where the last three elements are optional."
     (www 4)
     (aka 4 ?,)
     (score -1)
-    (x-face -1))
+    (x-face -1)
+    (face -1)
+    (sender -1))
   "Alist of entry types for presentation.
 The format of elements of this list should be
      (ENTRY SCORE [CLASS READ-ONLY])
@@ -149,14 +153,14 @@ The sender is passed to each function as the argument."
   :group 'lsdb
   :type 'hook)
 
-(defcustom lsdb-update-record-functions
+(defcustom lsdb-after-update-record-functions
   '(lsdb-update-address-cache)
   "List of functions called after a record is updated.
 The updated record is passed to each function as the argument."
   :group 'lsdb
   :type 'hook)
 
-(defcustom lsdb-delete-record-functions
+(defcustom lsdb-after-delete-record-functions
   '(lsdb-delete-address-cache)
   "List of functions called after a record is removed.
 The removed record is passed to each function as the argument."
@@ -201,7 +205,35 @@ The compressed face will be piped to this command."
   :group 'lsdb
   :type 'function)
 
-(defcustom lsdb-print-record-hook '(lsdb-expose-x-face)
+(defcustom lsdb-face-image-type nil
+  "A image type of displayed face.
+If non-nil, supersedes the return value of `lsdb-x-face-available-image-type'."
+  :group 'lsdb
+  :type 'symbol)
+
+(defcustom lsdb-face-command-alist
+  '((png "pngtopnm | pnmscale 0.5 | pnmtopng")
+    (xpm "pngtopnm | pnmscale 0.5 | ppmtoxpm"))
+  "An alist from an image type to a command to be executed to display a Face header.
+The command will be executed in a sub-shell asynchronously.
+The decoded field-body (actually a PNG data) will be piped to this command."
+  :group 'lsdb
+  :type 'list)
+
+(defcustom lsdb-insert-face-function
+  (if (static-if (featurep 'xemacs)
+	  (or (featurep 'png)
+	      (featurep 'xpm))
+	(and (>= emacs-major-version 21)
+	     (fboundp 'image-type-available-p)
+	     (or (image-type-available-p 'png)
+		 (image-type-available-p 'xpm))))
+      #'lsdb-insert-face-asynchronously)
+  "Function to display Face."
+  :group 'lsdb
+  :type 'function)
+
+(defcustom lsdb-print-record-hook '(lsdb-expose-x-face lsdb-expose-face)
   "A hook called after a record is displayed."
   :group 'lsdb
   :type 'hook)
@@ -249,6 +281,16 @@ Bourne shell or its equivalent \(not tcsh) is needed for \"2>\"."
   :type 'boolean
   :group 'lsdb)
 
+(defcustom lsdb-strip-address nil
+  "If non-nil, strip display-name from sender address before completion."
+  :group 'lsdb
+  :type 'boolean)
+
+(defcustom lsdb-use-migemo nil
+  "If non-nil, use `migemo' when complete address."
+  :type 'boolean
+  :group 'lsdb)
+
 ;;;_. Faces
 (defface lsdb-header-face
   '((t (:underline t)))
@@ -464,7 +506,8 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
 		       (symbol-name (coding-system-name
 				     lsdb-file-coding-system))))))
 	    (if coding-system-name
-		(insert ";;; -*- coding: " coding-system-name " -*-\n"))))
+		(insert ";;; -*- mode: emacs-lisp; coding: "
+			coding-system-name " -*-\n"))))
       (lsdb-insert-hash-table lsdb-hash-table)
       ;; Save the secondary hash tables following.
       (setq tables lsdb-secondary-hash-tables)
@@ -475,17 +518,16 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
 	(setq tables (cdr tables))))))
 
 ;;;_. Mail Header Extraction
-(defun lsdb-fetch-field-bodies (regexp)
+(defun lsdb-fetch-fields (regexp)
   (save-excursion
     (goto-char (point-min))
     (let ((case-fold-search t)
 	  field-bodies)
       (while (re-search-forward (concat "^\\(" regexp "\\):[ \t]*")
 				nil t)
-	(push (funcall lsdb-decode-field-body-function
-			     (buffer-substring (point) (std11-field-end))
-			     (match-string 1))
-		    field-bodies))
+	(push (cons (match-string 1)
+		    (buffer-substring (point) (std11-field-end)))
+	      field-bodies))
       (nreverse field-bodies))))
 
 (defun lsdb-canonicalize-spaces-and-dots (string)
@@ -526,14 +568,15 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
     (while tables
       (when (or force (not (symbol-value (car tables))))
 	(set (car tables) (lsdb-make-hash-table :test 'equal))
-	(lsdb-maphash
-	 (lambda (key value)
-	   (run-hook-with-args
-	    'lsdb-update-record-functions
-	    (cons key value)))
-	 lsdb-hash-table)
 	(setq lsdb-hash-tables-are-dirty t))
-      (setq tables (cdr tables)))))
+      (setq tables (cdr tables))))
+  (if lsdb-hash-tables-are-dirty
+      (lsdb-maphash
+       (lambda (key value)
+	 (run-hook-with-args
+	  'lsdb-after-update-record-functions
+	  (cons key value)))
+       lsdb-hash-table)))
 
 (defun lsdb-maybe-load-hash-tables ()
   (unless lsdb-hash-table
@@ -595,8 +638,10 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
 ;;;_ : Update Records
 (defun lsdb-update-record (sender &optional interesting)
   (let ((old (lsdb-gethash (car sender) lsdb-hash-table))
-	(new (cons (cons 'net (list (nth 1 sender)))
-		   interesting))
+	(new (if (nth 1 sender)
+		 (cons (cons 'net (list (nth 1 sender)))
+		       interesting)
+	       interesting))
 	merged
 	record
 	full-name)
@@ -625,7 +670,7 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
 			       (cdr record)))))
       (lsdb-puthash (car record) (cdr record)
 		    lsdb-hash-table)
-      (run-hook-with-args 'lsdb-update-record-functions record)
+      (run-hook-with-args 'lsdb-after-update-record-functions record)
       (setq lsdb-hash-tables-are-dirty t))
     record))
 
@@ -635,25 +680,46 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
     (save-restriction
       (std11-narrow-to-header)
       (setq senders
-	    (delq nil (mapcar #'lsdb-extract-address-components
-			      (lsdb-fetch-field-bodies
+	    (delq nil (mapcar (lambda (field)
+				(let ((components
+				       (lsdb-extract-address-components
+					(cdr field))))
+				  (if components
+				      (setcar
+				       components
+				       (funcall lsdb-decode-field-body-function
+						(car components) (car field))))
+				  components))
+			      (lsdb-fetch-fields
 			       lsdb-sender-headers)))
 	    recipients
-	    (delq nil (mapcar #'lsdb-extract-address-components
-			      (lsdb-fetch-field-bodies
+	    (delq nil (mapcar (lambda (field)
+				(let ((components
+				       (lsdb-extract-address-components
+					(cdr field))))
+				  (if components
+				      (setcar
+				       components
+				       (funcall lsdb-decode-field-body-function
+						(car components) (car field))))
+				  components))
+			      (lsdb-fetch-fields
 			       lsdb-recipients-headers))))
       (setq alist lsdb-interesting-header-alist)
       (while alist
 	(setq bodies
 	      (delq nil (mapcar
-			 (lambda (field-body)
-			   (if (nth 1 (car alist))
-			       (and (string-match (nth 1 (car alist))
-						  field-body)
-				    (replace-match (nth 3 (car alist))
-						   nil nil field-body))
-			     field-body))
-			 (lsdb-fetch-field-bodies (car (car alist))))))
+			 (lambda (field)
+			   (let ((field-body
+				  (funcall lsdb-decode-field-body-function
+					   (cdr field) (car field))))
+			     (if (nth 1 (car alist))
+				 (and (string-match (nth 1 (car alist))
+						    field-body)
+				      (replace-match (nth 3 (car alist))
+						     nil nil field-body))
+			       field-body)))
+			 (lsdb-fetch-fields (car (car alist))))))
 	(when bodies
 	  (setq entry (or (nth 2 (car alist))
 			  'notes))
@@ -796,12 +862,18 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
 ;;;_ : Matching Highlight
 (defvar lsdb-last-highlight-overlay nil)
 
+;;; avoid byte-compile warning for migemo
+(eval-when-compile
+  (autoload 'migemo-get-pattern "migemo"))
+
 (defun lsdb-complete-name-highlight (start end)
   (make-local-hook 'pre-command-hook)
   (add-hook 'pre-command-hook 'lsdb-complete-name-highlight-update nil t)
   (save-excursion
     (goto-char start)
-    (search-forward lsdb-last-completion end)
+    (if (and lsdb-use-migemo (fboundp 'migemo-get-pattern))
+	(re-search-forward lsdb-last-completion end)
+      (search-forward lsdb-last-completion end))
     (setq lsdb-last-highlight-overlay
 	  (make-overlay (match-beginning 0) (match-end 0)))
     (overlay-put lsdb-last-highlight-overlay 'face
@@ -836,23 +908,23 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
     (unless (eq last-command this-command)
       (setq lsdb-last-candidates nil
 	    lsdb-last-candidates-pointer nil
-	    lsdb-last-completion (buffer-substring start (point))
-	    pattern (concat "\\<" (regexp-quote lsdb-last-completion)))
+	    lsdb-last-completion (buffer-substring start (point)))
+      (if (and lsdb-use-migemo (fboundp 'migemo-get-pattern))
+	  (setq lsdb-last-completion (migemo-get-pattern lsdb-last-completion)
+		pattern (concat "\\<\\(" lsdb-last-completion "\\)"))
+	(setq pattern (concat "\\<" (regexp-quote lsdb-last-completion))))
       (lsdb-maphash
        (lambda (key value)
-	 (let ((net (cdr (assq 'net value))))
-	   (if (string-match pattern key)
-	       (setq lsdb-last-candidates
-		     (nconc lsdb-last-candidates
-			    (mapcar (lambda (address)
-				      (if (equal key address)
-					  key
-					(concat key " <" address ">")))
-				    net)))
-	     (while net
-	       (if (string-match pattern (car net))
-		   (push (car net) lsdb-last-candidates))
-	       (setq net (cdr net))))))
+	 (setq lsdb-last-candidates
+	       (nconc lsdb-last-candidates
+		      (delq nil (mapcar
+				 (lambda (candidate)
+				   (if (string-match pattern candidate)
+				       candidate))
+				 (if lsdb-strip-address
+				     (cdr (assq 'net value))
+				   (append (cdr (assq 'net value))
+					   (cdr (assq 'sender value)))))))))
        lsdb-hash-table)
       ;; Sort candidates by the position where the pattern occurred.
       (setq lsdb-last-candidates
@@ -1013,7 +1085,7 @@ Modify whole identification by side effect."
 (defun lsdb-delete-record (record)
   "Delete given RECORD."
   (lsdb-remhash (car record) lsdb-hash-table)
-  (run-hook-with-args 'lsdb-delete-record-functions record)
+  (run-hook-with-args 'lsdb-after-delete-record-functions record)
   (setq lsdb-hash-tables-are-dirty t))
 
 (defun lsdb-current-entry ()
@@ -1042,7 +1114,7 @@ Modify whole identification by side effect."
   (setcdr record (delq entry (cdr record)))
   (lsdb-puthash (car record) (cdr record)
 		lsdb-hash-table)
-  (run-hook-with-args 'lsdb-update-record-functions record)
+  (run-hook-with-args 'lsdb-after-update-record-functions record)
   (setq lsdb-hash-tables-are-dirty t))
 
 (defun lsdb-mode-add-entry (entry-name)
@@ -1069,7 +1141,7 @@ Modify whole identification by side effect."
 	      (setcdr record (cons (cons ',entry-name form) (cdr record)))
 	      (lsdb-puthash (car record) (cdr record)
 			    lsdb-hash-table)
-	      (run-hook-with-args 'lsdb-update-record-functions record)
+	      (run-hook-with-args 'lsdb-after-update-record-functions record)
 	      (setq lsdb-hash-tables-are-dirty t)
 	      (beginning-of-line 2)
 	      (add-text-properties
@@ -1156,7 +1228,7 @@ then the entire entry will be deleted."
 		 (entry (assq entry-name (cdr record))))
 	    (unless (equal form (cdr entry))
 	      (setcdr entry form)
-	      (run-hook-with-args 'lsdb-update-record-functions record)
+	      (run-hook-with-args 'lsdb-after-update-record-functions record)
 	      (setq lsdb-hash-tables-are-dirty t)
 	      (with-current-buffer lsdb-buffer-name
 		(let ((inhibit-read-only t)
@@ -1191,7 +1263,7 @@ then the entire entry will be deleted."
 	    (lsdb-delete-record record)
 	    (setcar record new-name)
 	    (lsdb-puthash new-name (cdr record) lsdb-hash-table)
-	    (run-hook-with-args 'lsdb-update-record-functions record)
+	    (run-hook-with-args 'lsdb-after-update-record-functions record)
 	    (setq lsdb-hash-tables-are-dirty t)
 	    (with-current-buffer lsdb-buffer-name
 	      (let ((inhibit-read-only t)
@@ -1583,7 +1655,7 @@ always hide."
 				    (cdr (car records))))
 	(lsdb-puthash (car (car records)) (cdr (car records))
 		      lsdb-hash-table)
-	(run-hook-with-args 'lsdb-update-record-functions (car records))
+	(run-hook-with-args 'lsdb-after-update-record-functions (car records))
 	(setq lsdb-hash-tables-are-dirty t)))))
 
 (defun lsdb-mu-get-prefix-method ()
@@ -1742,20 +1814,114 @@ the user wants it."
 				lsdb-x-face-cache)))
 	    (kill-buffer (process-buffer process))))))))
 
+;;;_. Face Rendering
+(defvar lsdb-face-cache
+  (lsdb-make-hash-table :test 'equal))
+
+(defun lsdb-face-available-image-type ()
+  (static-if (featurep 'xemacs)
+      (if (featurep 'png)
+	  'png
+	(if (featurep 'xpm)
+	    'xpm))
+    (and (>= emacs-major-version 21)
+	 (fboundp 'image-type-available-p)
+	 (if (image-type-available-p 'png)
+	     'png
+	   (if (image-type-available-p 'xpm)
+	       'xpm)))))
+
+(defun lsdb-expose-face ()
+  (let* ((record (get-text-property (point-min) 'lsdb-record))
+	 (face (cdr (assq 'face (cdr record))))
+	 (delimiter "\r "))
+    (when (and lsdb-insert-face-function
+	       face)
+      (goto-char (point-min))
+      (end-of-line)
+      (put-text-property 0 1 'invisible t delimiter) ;hide "\r"
+      (put-text-property
+       (point)
+       (progn
+	 (insert delimiter)
+	 (while face
+	   (funcall lsdb-insert-face-function (pop face)))
+	 (point))
+       'lsdb-record record))))
+
+(defun lsdb-insert-face-image (data type marker)
+  (static-if (featurep 'xemacs)
+      (save-excursion
+	(set-buffer (marker-buffer marker))
+	(goto-char marker)
+	(let* ((inhibit-read-only t)
+	       buffer-read-only
+	       (glyph (make-glyph (vector type :data data))))
+	  (set-extent-begin-glyph
+	   (make-extent (point) (point))
+	   glyph)))
+    (save-excursion
+      (set-buffer (marker-buffer marker))
+      (goto-char marker)
+      (let* ((inhibit-read-only t)
+	     buffer-read-only
+	     (image (create-image data type t :ascent 'center))
+	     (record (get-text-property (point) 'lsdb-record)))
+	(put-text-property (point) (progn
+				     (insert-image image)
+				     (point))
+			   'lsdb-record record)))))
+
+(defun lsdb-insert-face-asynchronously (face)
+  (let* ((type (or lsdb-face-image-type
+		   (lsdb-face-available-image-type)))
+	 (shell-file-name lsdb-shell-file-name)
+	 (shell-command-switch lsdb-shell-command-switch)
+	 (coding-system-for-read 'binary)
+	 (coding-system-for-write 'binary)
+	 (process-connection-type nil)
+	 (cached (cdr (assq type (lsdb-gethash face lsdb-face-cache))))
+	 (marker (point-marker))
+	 process)
+    (if cached
+	(lsdb-insert-face-image cached type marker)
+      (setq process
+	    (start-process-shell-command
+	     "lsdb-face-command" (generate-new-buffer " *lsdb work*")
+	     (concat "{ "
+		     (nth 1 (assq type lsdb-face-command-alist))
+		     "; } 2> /dev/null")))
+      (process-send-string process (base64-decode-string face))
+      (process-send-eof process)
+      (set-process-sentinel
+       process
+       `(lambda (process string)
+	  (unwind-protect
+	      (when (and (buffer-live-p (marker-buffer ,marker))
+			 (equal string "finished\n"))
+		(let ((data
+		       (with-current-buffer (process-buffer process)
+			 (set-buffer-multibyte nil)
+			 (buffer-string))))
+		  (lsdb-insert-face-image data ',type ,marker)
+		  (lsdb-puthash ,face (list (cons ',type data))
+				lsdb-face-cache)))
+	    (kill-buffer (process-buffer process))))))))
+
 (require 'product)
 (provide 'lsdb)
 
 (product-provide 'lsdb
-  (product-define "LSDB" nil '(0 9)))
+  (product-define "LSDB" nil '(0 11)))
 
 ;;;_* Local emacs vars.
-;;; The following `outline-layout' local variable setting:
+;;; The following `allout-layout' local variable setting:
 ;;;  - closes all topics from the first topic to just before the third-to-last,
 ;;;  - shows the children of the third to last (config vars)
 ;;;  - and the second to last (code section),
 ;;;  - and closes the last topic (this local-variables section).
 ;;;Local variables:
-;;;outline-layout: (0 : -1 -1 0)
+;;;allout-layout: (0 : -1 -1 0)
 ;;;End:
 
 ;;; lsdb.el ends here
